!!$=head1 NAME
!!$
!!$ddot - calculate double precission dot product
!!$
!!$=head1 SYNOPSIS
!!$
!!$C<product=ddot(arr1,arr2)>
!!$
!!$=head1 INPUT PARAMETERS
!!$
!!$=over 4
!!$
!!$=item arr1,arr2 - C<real(:)>  
!!$
!!$      arrays to calcuate dot products of
!!$
!!$=back   
!!$
!!$=head1 RETURN VALUE
!!$
!!$=over 4
!!$
!!$=item product  - double precission 
!!$
!!$      result of dot product
!!$
!!$=back
!!$
!!$=head1  DESCRIPTION 
!!$
!!$Calculate a dout product of two real vectors.
!!$
!!$=head1 LIBRARY
!!$
!!$B<geef90>
!!$
!!$
!!$=cut
!!$

module ddot_mod
 use sep_mod
  implicit none
contains
  function ddot (arr1, arr2) result (product)
    double precision 			:: product
    real, dimension(:), intent(in) 	:: arr1, arr2
    product = sum(dprod(arr1, arr2))
  end function ddot
  double precision function dot_product_d_r(a1,a2) result(x)
    real :: a1(:)
    real :: a2(:)
    integer :: i
    x=0
    !$OMP PARALLEL DO private(i) reduction(+:x)
    do i=1,size(a1)
      x=x+a1(i)*a2(i)
    end do
    !$OMP END PARALLEL DO
  end function
  double precision function dot_product2(a1,a2) result(x)
    real :: a1(:),a2(:)
    integer :: i
    x=0
    !$OMP PARALLEL DO private(i) reduction(+:x)
    do i=1,size(a1)
      x=x+a1(i)*a2(i)
    end do
    !$OMP END PARALLEL DO
  end function
  subroutine dprod_r_r_d(a1,a2,b)
   real :: a1(:),a2(:)
   real :: b(:)
   integer :: i
   !$OMP PARALLEL DO private(i)
   do i=1,size(b)
     b(i)=a1(i)*a2(i) 
   end do
   !$OMP END PARALLEL DO
   end subroutine
      
   
    
   
end module ddot_mod


