module pursuit_mod
 
use sep_mod
contains
  !     -------------------------------------------------------------------
  !!     Minimize c'x +gamma^2x'x+p'p
  !!     subject Ax+delta.p=b, x.ge.0
  !!     (following Chen et all.)
  subroutine solve_pursuit       (op,x,b,iterlp,itercg,tol,delta,gamma,mu,&
     PI,DI,DG)
    use sep
    implicit none 
    interface 
      integer function op(adj,add,mod,dat)
        real :: mod(:),dat(:)
        logical,intent(in) :: adj,add
      end function
    end interface
    integer                ::nm,nd,  na,                iterlp,itercg
    real   ,dimension(:)  ::           x,b;
    real,allocatable,dimension(:) :: jj
    real            ::                                                tol,gamma,mu,PI,DI,DG
    real,allocatable :: c(:)
    real,allocatable,dimension(:):: D,v,t,r,z,y,rhs,xp,yp,dx,dy,dz
    real                         :: delta,rhod,rhop,x_one,z_one,PI_t,DI_t,DG_t
    integer :: itmax,k,i,kz,kx,ierr
    real    :: step_p,step_d


    nm=size(x)
    nd=size(b)
    allocate(c(nm))
    c=1
    itmax=iterlp
    PI_t=PI       ! _t will be the comparisson standards
    DI_t=DI 
    DG_t=DG

    !! Set inital variables (D is not the data- it's a big diagonal matrix
    allocate( t(nm), v(nm),D(nm),z(nm), xp(nm)); t=0.; v=0.;d=0.;z=1.; xp=0.
    allocate(yp(nd),dy(nd),y(nd),r(nd),rhs(nd),jj(nm/2));yp=0.;dy=0.;y=1.;r=0.;rhs=0.

    iterlp=0
    do k=1,itmax
       write(0,*) ' '
       write(0,*) ' -------------------------------------------'
       write(0,*) '    Primal-Dual iteration',k
       iterlp=iterlp+1
       !!  Form least squares system to solve
       !! Develop LS equation for Interior point search direction
       ierr=oper2(.true.,.false.,xp,y,op)!xp=A'y
     !  call  mat_mul (1,a,na,nm,xp,nd,y,irow,icol) !!xp=A'y
       !$OMP PARALLEL DO private(i)
       do i=1,size(t)
         t(i)  =c(i)   +gamma*gamma*x(i)   -z(i)   -xp(i)
         v(i)   =mu-x(i)   *z(i)
       end do
       !$OMP END PARALLEL DO
       !$OMP PARALLEL DO private(i)
       do i=1,size(d)
         d(i)   =1./(z(i)   /x(i)   +gamma*gamma)
       end do
       !$OMP END PARALLEL DO
       ierr=oper2(.false.,.false.,x,yp,op)
      ! call  mat_mul (0,a,na,nm,x,nd,yp,irow,icol)!!yp=Ax
       !$OMP WORKSHARE
       r   =b   -yp   -delta*delta*y
       !$OMP END WORKSHARE
       !!     right side term of the system
       !$OMP WORKSHARE
       xp   =d   *(v   /x   -t   )
       !$OMP END WORKSHARE

       ierr=oper2(.false.,.false.,xp,yp,op)
      ! call  mat_mul (0,a,na,nm,xp,nd,yp,irow,icol) !!yp=Axp
       rhs   =r   -yp

       !!     solve ADA'dy=rhs least squares system for dual variable update (dy)
       call CG_solver(op,D,nm,dy,rhs,delta,itercg,tol)

       !!    set dx and dz from dy answer
       ierr=oper2(.true.,.false.,xp,dy,op)!!xp=A'dy
       allocate(dx(nm),dz(nm));dx=0.;dz=0.
       dx   =d   *xp   +d   *(v   /x   -t   )
       dz   =v   /x   -z   *dx   /x   

       !!    primal and dual step sizes        
       x_one=x(1)+dx(1)
       z_one=z(1)+dz(1)
       kx=1
       kz=1
       do i=2,nm
          if(z_one.ge.(z(i)+dz(i))) then
             z_one=z(i)+dz(i)
             kz=i
          endif
          if(x_one.ge.(x(i)+dx(i))) then 
             x_one=x(i)+dx(i)
             kx=i
          endif
       enddo

       if(x_one.ge.0.) step_p=1.
       if(x_one.lt.0.) step_p=-x(kx)/dx(kx)

       if(z_one.ge.0.) step_d=1.
       if(z_one.lt.0.) step_d=-z(kz)/dz(kz)

       rhop=.99*step_p
       rhod=.99*step_d

      !$OMP PARALLEL
      !$OMP WORKSHARE
       x   =x   +rhop*dx   
       z   =z   +rhod*dz
       y   =y   +rhod*dy   

       !$OMP END WORKSHARE
       !$OMP END PARALLEL
       mu=(1.- min(rhop,rhod,0.99))*mu
       write(0,*) 'mu should be approaching zero:',mu  

       !!       primal-dual (convergence) condition test 
       PI=dot(nd,r,r)/(dot(nm,x,x)+1.0)
       DI=dot(nm,t,t)/(dot(nd,y,y)+1.0)
       DG=dot(nm,z,x)/(dot(nm,x,x)+1.0)
       write(0,*) ' '
       write(0,*) 'Current Daulity Gap =',DG
       write(0,*) '             DG goal=',DG_t
       if(PI.lt.PI_t.and.DI.lt.DI_t .and. DG.lt.DG_t) then
          write(0,*) 'Converged to feasability and duality gap parameters.'
          return
       end if
       deallocate(dz,dx)
       jj=x(1:nm/2)-x(nm/2+1:)
       ierr= srite("xmod.H",jj,size(jj)*4)
       allocate(dz(nd))
              ierr=oper2(.false.,.false.,x,dz,op)!!xp=A'dy

       ierr= srite("rr.H",dz,size(dz)*4)
       deallocate(dz)
    end do !main iteration loop
       deallocate(t,v,D,z,xp,yp,dy,y,r,rhs,c)
        deallocate(jj)
  end subroutine

  integer function oper2(adj,add,model,data,op)
    interface
      integer function op(adj,add,model,data)
        logical,intent(in) :: adj,add
        real :: model(:),data(:)
      end function
    end interface
    logical,intent(in) :: adj,add
    real :: model(:),data(:)
    real,allocatable :: temp(:)
    allocate(temp(size(model)/2))
    oper2=0
    if(.not. adj) then
      if(.not. add) data=0
      oper2=op(.false.,.false.,model(:size(model)/2),data)
      temp=-model(size(model)/2+1:)
      oper2=op(.false.,.true.,temp,data)
    else
      if(.not. add) model=0
      oper2=op(.true.,.false.,temp,data)
      model(1:size(model)/2)=temp+model(:size(model)/2)
      model(1+size(model)/2+1:)=-temp+model(size(model)/2+1:)
    end if
    deallocate(temp)
  end function

  !!        -------------------------------------

 double precision function dot( n, x, y)
    implicit none
    real,dimension(:) ::x, y
    integer:: n,i
    real   ::val 
    dot=-999
    val=0.0

    do i=1,n
       val = val + x(i)*y(i)
    enddo

    dot=val

    return
  end function dot
  !!     ----------------------------------------
  !!    Sparse matrix multiplication when operators
  !!    can be stored in sparse matrix format
  !!    maybe a bit more efficient than operators

  !!---------------------------------------------------------------
  !!     Solve the system
  !!                        ADA'x=b when A is sparse and D is
  !!                        diagonal      
  !!     use in each iteration of PD_lp_bm - 
  subroutine  CG_solver(op,D,nm,x,b,delta,itercg,tol)
    implicit none
    interface
      integer function op(adj,add,mod,dat)
        real :: mod(:),dat(:)
        logical,intent(in) :: adj,add
      end function
    end interface
    integer        ::       na,nm,nd,                    itercg,ierr
    real           ::                              delta,       tol
    real ,dimension(:)::D,                     x,b
    real           :: beta,rho,rho_old, alpha
    integer        :: iter_max,countCG,niter,iter
    integer :: i
    real ,allocatable,dimension(:)  ::  r,p,w,xp

   write(0,*) "SUM IN ",sum(D),sum(x),sum(b)
    nd=size(x)
    allocate(r(nd),p(nd),w(nd) )
    allocate(xp(nm))
    !$OMP PARALLEL DO private(i)
    do i=1,size(x)
      x(i)=0. !! first model
    end do
    !$OMP END PARALLEL DO
    !$OMP PARALLEL DO private(i)
    do i=1,size(r)
      r(i)=b(i)  !! first residual
    end do
    !$OMP END PARALLEL DO
    rho=dot(nd,r,r)
    countCG=0
    write(0,*) "ITER CG",itercg,iter_max
    niter=itercg

    do iter=1,niter
       countCG=countCG+1
       if(iter.eq.1) then
         !$OMP PARALLEL DO private(i)
         do i=1,size(p)
            p(i)=r(i)
         end do
         !$OMP END PARALLEL DO
       else
          beta=rho/rho_old
          !$OMP PARALLEL DO
          do i=1,size(p)
            p(i)=r(i)+beta*p(i)
          end do
          !$OMP END PARALLEL DO
       endif
       !!        xp=A'p
       ierr=oper2(.true.,.false.,xp,p,op)
       !!        D.xp
       !$OMP PARALLEL DO private(i)
       do i=1,size(xp)
         xp(i)=D(i)*xp(i)
       end do
       !$OMP END PARALLEL DO
       ierr=oper2(.false.,.false.,xp,w,op)
       !!       Regularization
       !$OMP PARALLEL DO private(i)
       do i=1,size(w)
         w(i)=w(i)+p(i)*delta**2
       end do
       !$OMP END PARALLEL DO
       alpha=rho/dot(nd,p,w)
       !$OMP PARALLEL DO private(i)
       do i=1,size(x)
         x(i)=x(i)+alpha*p(i)
        end do
       !$OMP END PARALLEL DO
       !$OMP PARALLEL DO private(i)
       do i=1,size(r)
         r(i)=r(i)-alpha*w(i)
        end do
       !$OMP END PARALLEL DO
       rho_old=rho
       rho=dot(nd,r,r)
       write(0,*) iter,"=iter gap=",sqrt(rho/nd)
       if(sqrt(rho/nd).le.tol) then
          write(0,*) ' '
          write(0,*)'CG iterations converged below tolerance of',tol 
          write(0,*) countCG,'iterations required.'
          write(0,*) ' '
          deallocate(r,p,w,xp)
          return
       end if
        
    enddo  ! main loop
    deallocate(r,p,w,xp)
    write(0,*) ' '
    write(0,*) "CG iterations insufficient to reach tolerance",tol
    write(0,*) "              stopped at ",sqrt(rho/nd)
    write(0,*) countCG,' CG iterations used for barrier step.'
    write(0,*) ' '
  end subroutine CG_solver

  
end module 
