module wavelet_mod
   use adj_mod
  use sep_mod
   implicit none
   
   real, private :: h0,h1,h2,h3,g0,g1,g2,g3
   real,private  :: ih0,ih1,ih2,ih3,ig0,ig1,ig2,ig3
   integer,private :: nt,nn1,nn2,nn3,levit
   real, private, pointer :: temp1(:),temp2(:)
   contains
   subroutine build_filter(n1,n2,n3,level)
     integer :: n1,n2,n3,level
     double precision  :: sqrt_3 = sqrt( 3. );
     double precision  :: denom = 4 * sqrt( 2. );

      levit=level
      nn1=n1; nn2=n2; nn3=n3
      nt=max(n1,max(n2,n3))
      allocate(temp1(nt),temp2(nt))
      //
      // forward transform scaling (smoothing) coefficients
      //
      h0 = (1 + sqrt_3)/denom;
      h1 = (3 + sqrt_3)/denom;
      h2 = (3 - sqrt_3)/denom;
      h3 = (1 - sqrt_3)/denom;
      //
      // forward transform wavelet coefficients
      //
      g0 =  h3;
      g1 = -h2;
      g2 =  h1;
      g3 = -h0;

      Ih0 = h2;
      Ih1 = g2;  // h1
      Ih2 = h0;
      Ih3 = g0;  // h3

      Ig0 = h3;
      Ig1 = g3;  // -h0
      Ig2 = h1;
      Ig3 = g1;  // -h2
      
    end subroutine
    subroutine forward_1d(in,out)
      integer :: i,n,half 
      real :: in(:),out(:)
      n=size(in)
      temp1=in
      do while(n >=4 )
        half=n/2
        call one_level_for_1d(temp1,temp2(1:half),temp2(half+1:n),n)
        out(half+1:n)=temp2(half+1:n)
        temp1(1:half)=temp2(1:half)
        n=n/2
      end do
      out(1:half)=temp2(1:half)
    end subroutine

    
    subroutine inverse_1d(in,out)
      integer :: i,n,half
      real    :: in(:),out(:)
      n=4
      temp1=in
      do while(n /= size(in))
        half=n/2
        call one_level_inv_1d(temp2,temp1(1:half),temp1(half+1:n),n)
        temp1(1:n)=temp2(1:n)
        n=n*2
      end do
      out=temp1
    end subroutine


  integer function wavelet_op(adj,add,model,data)
    logical, intent(in) :: adj,add
    real :: model(:),data(:)
    real,pointer :: model3(:,:,:),data3(:,:,:)
    call adjnull(adj,add,model,data)
    allocate(model3(nn1,nn2,nn3),data3(nn1,nn2,nn3))
    model3=reshape(model,(/nn1,nn2,nn3/))
    data3=reshape(data,(/nn1,nn2,nn3/))
    if(adj) then
       call forward_3d(data3,model3)
    else
       call inverse_3d(model3,data3)
    end if
    model=reshape(model3,(/size(model)/))
    data=reshape(data3,(/size(data)/))
    deallocate(model3,data3)
    wavelet_op=0
  end function

   subroutine forward_3d(in,out)
     real, pointer :: in(:,:,:),out(:,:,:)
     real,pointer :: temp1(:,:,:),temp2(:,:,:)
     integer      :: n1,n2,n3,nn1,nn2,nn3,lev1,lev2,lev3,lev
     lev1=calc_level(size(in,1))
     lev2=calc_level(size(in,2))
     lev3=calc_level(size(in,3))
     n1=size(in,1); n2=size(in,2); n3=size(in,3)

     allocate(temp1(size(in,1),size(in,2),size(in,3)))
     allocate(temp2(size(in,1),size(in,2),size(in,3)))
     lev=levit
     !write(0,*) "TEST",lev
     temp1=in
     do while(lev/=0)
       call one_level_3d_forward(temp1,temp2,n1,n2,n3)
       temp1(1:n1,1:n2,1:n3)=temp2(1:n1,1:n2,1:n3)
       if(n1>=4) n1=n1/2
       if(n2>=4) n2=n2/2
       if(n3>=4) n3=n3/2
       lev=lev-1
     end do
     out=temp1+out
     deallocate(temp1,temp2)
   end subroutine
   
   subroutine inverse_3d(in,out)
     real,pointer :: in(:,:,:),out(:,:,:)
     real,pointer :: temp1(:,:,:),temp2(:,:,:)
     integer      :: n1,n2,n3,nn1,nn2,nn3,lev1,lev2,lev3,lev
     lev1=calc_level(size(in,1))
     lev2=calc_level(size(in,2))
     lev3=calc_level(size(in,3))
     lev=levit
     n1=2; n2=2; n3=2
     n1=size(in,1)/2**(lev-1)
     n2=size(in,2)/2**(lev-1)
     n3=size(in,3)/2**(lev-1)
     if(n1<2) n1=2
     if(n2<2) n2=2
     if(n3<2) n3=2
     if(size(in,1)==1) n1=1
     if(size(in,2)==1) n2=1
     if(size(in,3)==1) n3=1
   write(0,*) "BBB",n1,n2,n3,lev
     allocate(temp1(size(in,1),size(in,2),size(in,3)))
     allocate(temp2(size(in,1),size(in,2),size(in,3)))

   !  temp1(1:n1,1:n2,1:n3)=in
     temp1=in
     do while(lev/=0)
       if(lev1==lev) n1=4
       if(lev2==lev) n2=4
       if(lev3==lev) n3=4
       temp2=0
       call one_level_3d_inverse(temp1,temp2,n1,n2,n3)
       temp1(1:n1,1:n2,1:n3)=temp2(1:n1,1:n2,1:n3)
       if(lev1>=lev) n1=n1*2
       if(lev2>=lev) n2=n2*2
       if(lev3>=lev) n3=n3*2
       lev=lev-1
     end do
     out=temp1+out
     deallocate(temp1,temp2)
   end subroutine
   integer function calc_level(n1)
     integer :: n1,j1,lev,nn
     
     nn=n1
     lev=0
     do while(nn >=4)
       nn=nn/2; lev=lev+1
     end do
     if(2**(1+lev)/=n1 .and. n1/=1) then
       write(0,*) nn,n1,lev
       call seperr("axes must be power of 2")
     end if
     calc_level=lev
   end function
     
     

  
    subroutine one_level_3d_inverse(in,out,n1,n2,n3)
    real,pointer :: in(:,:,:),out(:,:,:)
    real,pointer,dimension(:,:,:) :: lll,llh,lhl,lhh,hll,hlh,hhl,hhh
    real,pointer,dimension(:,:,:) :: temp1,temp2,temp3,temp4
    integer :: n1,n2,n3

    call set_3d_pointers(in,lll,llh,lhl,lhh,hll,hlh,hhl,hhh,n1,n2,n3)
    if(n1>4) then
      if(n2>4) then
        if(n3>4) then
           allocate(temp1(n1/2,n2,n3),temp2(n1/2,n2,n3))
           allocate(temp3(n1/2,n2/2,n3),temp4(n1/2,n2/2,n3))
           call one_level_3d_axis_inv(temp3,hll,hlh,3,n1/2,n2/2,n3)   !HLL HLH
           call one_level_3d_axis_inv(temp4,hhl,hhh,3,n1/2,n2/2,n3)   !HHL HHH      
           call one_level_3d_axis_inv(temp2,temp3,temp4,2,n1/2,n2,n3) !HLX HHX
           call one_level_3d_axis_inv(temp3,lll,llh,3,n1/2,n2/2,n3)   !LLL LLH
           call one_level_3d_axis_inv(temp4,lhl,lhh,3,n1/2,n2/2,n3)   !LHL LHH
           call one_level_3d_axis_inv(out   ,temp1,temp2,1,n1,n2,n3) !LXX, HXX 
           call one_level_3d_axis_inv(temp1,temp3,temp4,2,n1/2,n2,n3) !LLX, LHX
           deallocate(temp2,temp1,temp3,temp4)
        else
           allocate(temp1(n1/2,n2,n3),temp2(n1/2,n2,n3))
           call one_level_3d_axis_inv(temp1,lll,lhl,2,n1/2,n2,n3)   !LLL, LHL
           call one_level_3d_axis_inv(temp2,hll,hhl,2,n1/2,n2,n3)   !HLL, HHL
  
   !       temp1=in(1:n1/2,:n2,:n3)
    !      temp2=in(n1/2+1:n1,:n2,:n3)

           call one_level_3d_axis_inv(out   ,temp1,temp2,1,n1,n2,n3) !LXX, HXX 

           deallocate(temp1,temp2)
        end if
      else 
        if(n3>4) then
          allocate(temp1(n1/2,n2,n3),temp2(n1/2,n2,n3))
          call one_level_3d_axis_inv(temp1,lll,llh,1,n1/2,n2,n3)
          call one_level_3d_axis_inv(temp2,hll,hlh,1,n1/2,n2,n3)
          call one_level_3d_axis_inv(out, temp1,temp2,1,n1,n2,n3)

          deallocate(temp1,temp2)
        else
          call one_level_3d_axis_inv(in,lll,hll,1,n1,n2,n3)
        end if
      end if
    else
      if(n2>4) then 
        if(n3>4) then
          allocate(temp1(n1,n2/2,n3),temp2(n1,n2/2,n3))
          call one_level_3d_axis_inv(temp1,lll,llh,1,n1,n2/2,n3)
          call one_level_3d_axis_inv(temp2,lhl,lhh,1,n1,n2/2,n3)
          call one_level_3d_axis_inv(out,temp1,temp2,1,n1,n2,n3)

          deallocate(temp1,temp2)
        else
          call one_level_3d_axis_inv(out,lll,lhl,1,n1,n2,n3)
        end if
      else
        if(n3>4) then
          call one_level_3d_axis_inv(out,lll,llh,1,n1,n2,n3)
        else
          lll=out
        end if
      end if
    end if

  end subroutine
  
  subroutine one_level_3d_forward(in,out,n1,n2,n3)
    real,pointer :: in(:,:,:),out(:,:,:)
    real,pointer,dimension(:,:,:) :: lll,llh,lhl,lhh,hll,hlh,hhl,hhh
    real,pointer,dimension(:,:,:) :: temp1,temp2,temp3,temp4
    integer :: n1,n2,n3,nh1,nh2,nh3,i1,i2,i3

    call set_3d_pointers(out,lll,llh,lhl,lhh,hll,hlh,hhl,hhh,n1,n2,n3)
    if(n1>4) then
      if(n2>4) then
        if(n3>4) then
           allocate(temp1(n1/2,n2,n3),temp2(n1/2,n2,n3))
           allocate(temp3(n1/2,n2/2,n3),temp4(n1/2,n2/2,n3))
           call one_level_3d_axis(in   ,temp1,temp2,1,n1,n2,n3) !LXX, HXX 
           call one_level_3d_axis(temp1,temp3,temp4,2,n1/2,n2,n3) !LLX, LHX
           call one_level_3d_axis(temp3,lll,llh,3,n1/2,n2/2,n3)   !LLL LLH
           call one_level_3d_axis(temp4,lhl,lhh,3,n1/2,n2/2,n3)   !LHL LHH
           call one_level_3d_axis(temp2,temp3,temp4,2,n1/2,n2,n3) !HLX HHX
           call one_level_3d_axis(temp3,hll,hlh,3,n1/2,n2/2,n3)   !HLL HLH
           call one_level_3d_axis(temp4,hhl,hhh,3,n1/2,n2/2,n3)   !HHL HHH      
           deallocate(temp2,temp1,temp3,temp4)
        else
           allocate(temp1(n1/2,n2,n3),temp2(n1/2,n2,n3))
           call one_level_3d_axis(in   ,temp1,temp2,1,n1,n2,n3) !LXX, HXX 
           nh1=n1/2; nh2=n2/2; nh3=n3/2
           call one_level_3d_axis(temp1,lll,lhl,2,n1/2,n2,n3)   !LLL, LHL
           call one_level_3d_axis(temp2,hll,hhl,2,n1/2,n2,n3)   !HLL, HHL
           deallocate(temp1,temp2)
        end if
      else 
        if(n3>4) then
          allocate(temp1(n1/2,n2,n3),temp2(n1/2,n2,n3))
          call one_level_3d_axis(in, temp1,temp2,1,n1,n2,n3)
  
          call one_level_3d_axis(temp1,lll,llh,1,n1/2,n2,n3)
          call one_level_3d_axis(temp2,hll,hlh,1,n1/2,n2,n3)
          deallocate(temp1,temp2)
        else
          call one_level_3d_axis(in,lll,hll,1,n1,n2,n3)
        end if
      end if
    else
      if(n2>4) then 
        if(n3>4) then
          allocate(temp1(n1,n2/2,n3),temp2(n1,n2/2,n3))
          call one_level_3d_axis(in,temp1,temp2,1,n1,n2,n3)
          call one_level_3d_axis(temp1,lll,llh,1,n1,n2/2,n3)
          call one_level_3d_axis(temp2,lhl,lhh,1,n1,n2/2,n3)
          deallocate(temp1,temp2)
        else
          call one_level_3d_axis(in,lll,lhl,1,n1,n2,n3)
        end if
      else
        if(n3>4) then
          call one_level_3d_axis(in,lll,llh,1,n1,n2,n3)
        else
          lll=in
        end if
      end if
    end if
  end subroutine
   subroutine one_level_3d_axis_inv(out,low,high,iax,n1,n2,n3)
      real :: out(:,:,:),low(:,:,:),high(:,:,:)
      integer :: n1,n2,n3,i1,i2,i3,iax

      if(iax==1)   then   
        do i3=1,n3
          do i2=1,n2
     !     write(0,*) "IN RTHI"
            call one_level_inv_1d(out(1:n1,i2,i3),low(1:n1/2,i2,i3),high(1:n1/2,i2,i3),n1)
          end do
        end do
      else if(iax==2) then
        do i3=1,n3
          do i1=1,n1
            call one_level_inv_1d(out(i1,1:n2,i3),low(i1,1:n2/2,i3),high(i1,1:n2/2,i3),n2)
          end do
        end do
      else
        do i2=1,n2
          do i1=1,n1
            call one_level_inv_1d(out(i1,i2,1:n3),low(i1,i2,1:n3/2),high(i1,i2,1:n3/2),n3)
          end do
        end do
      end if
    end subroutine

  
  
  subroutine one_level_3d_axis(in,low,high,iax,n1,n2,n3)
      real :: in(:,:,:),low(:,:,:),high(:,:,:)
      integer :: n1,n2,n3,i1,i2,i3,iax

      if(iax==1)   then   
        do i3=1,n3
          do i2=1,n2
            call one_level_for_1d(in(:n1,i2,i3),low(:n1/2,i2,i3),high(:n1/2,i2,i3),n1)
          end do
        end do
      else if(iax==2) then
        do i3=1,n3
          do i1=1,n1
            call one_level_for_1d(in(i1,1:n2,i3),low(i1,1:n2/2,i3),high(i1,1:n2/2,i3),n2)
          end do
        end do
      else
        do i2=1,n2
          do i1=1,n1
            call one_level_for_1d(in(i1,i2,1:n3),low(i1,i2,1:n3/2),high(i1,i2,1:n3/2),n3)
          end do
        end do
      end if
    end subroutine
  subroutine set_3d_pointers(buf,lll,llh,lhl,lhh,hll,hlh,hhl,hhh,n1,n2,n3)
    real,pointer,dimension(:,:,:) :: lll,llh,lhl,lhh,hll,hlh,hhl,hhh,buf
    integer :: n1,n2,n3
    
    if(n1>=4) then!{
      if(n2>=4) then!{
        if(n3>=4) then!{
          lll=>buf(:n1/2,:n2/2,:n3/2); llh=>buf(:n1/2,:n2/2,n3/2+1:n3)
          lhl=>buf(:n1/2,n2/2+1:n2,:n3/2); lhh=>buf(:n1/2,n2/2+1:,n3/2+1:n3)
          hll=>buf(n1/2+1:n1,:n2/2,:n3/2); hlh=>buf(n1/2+1:,:n2/2,n3/2+1:n3)
          hhl=>buf(n1/2+1:n1,n2/2+1:n2,:n3/2); hhh=>buf(n1/2+1:,n2/2+1:,n3/2+1:n3)
        !}
        else!{
          lll=>buf(:n1/2,:n2/2,:n3); 
          lhl=>buf(:n1/2,n2/2+1:n2,:n3); 
          hll=>buf(n1/2+1:n1,:n2/2,:n3);
          hhl=>buf(n1/2+1:n1,n2/2+1:n2,:n3); 
        end if!}
      !}
      else!{
        if(n3>=4) then!{
          lll=>buf(:n1/2,:n2,:n3/2); llh=>buf(:n1/2,:n2/2,n3/2+1:n3)
          hll=>buf(n1/2+1:n1,:n2,:n3/2); hlh=>buf(n1/2+1:,:n2,n3/2+1:n3)
        !}
        else!{
          lll=>buf(:n1/2,:n2,:n3); 
          hll=>buf(n1/2+1:n1,:n2,:n3);
        !}
        end if
      !}
      end if
    !}
    else     !{
      if(n2>=4) then!{
        if(n3>=4) then!{
          lll=>buf(:n1,:n2/2,:n3/2); llh=>buf(:n1,:n2/2,n3/2+1:n3)
          lhl=>buf(:n1,n2/2+1:n2,:n3/2); lhh=>buf(:n1,n2/2+1:,n3/2+1:n3)
        !}
        else!{
          lll=>buf(:n1,:n2/2,:n3); 
          lhl=>buf(:n1,n2/2+1:n2,:n3); 
        end if!}
      !}
      else!{
        if(n3>=4) then!{
          lll=>buf(:n1,:n2,:n3/2); llh=>buf(:n1,:n2,n3/2+1:n3)
        !}
        else!{
          lll=>buf(:n1,:n2,:n3); 
        end if!}
     end if!}
    !}
    end if
  end subroutine
  
   subroutine one_level_for_1d(in,low,high,n)
      real     :: in(:),low(:),high(:)
      integer  :: i,j,half,n
      
      half=size(in)/2
      i=1; j=1;
      do while(j<=n-3)
        low(i)=in(j)*h0+in(j+1)*h1+in(j+2)*h2+in(j+3)*h3
        high(i)=in(j)*g0+in(j+1)*g1+in(j+2)*g2+in(j+3)*g3
        j=j+2
        i=i+1
      end do
      low(i)=in(n-1)*h0+in(n)*h1+in(1)*h2+in(2)*h3
      high(i)=in(n-1)*g0+in(n)*g1+in(1)*g2+g3*in(2)
    end subroutine
    


    subroutine one_level_inv_1d(out,low,high,n)
      real  :: low(:),high(:),out(:)
      integer :: i,n,half,hp,j
      
      half=n/2
      hp=half+1
      
      out(1)=low(half)*ih0+high(half)*ih1+low(1)*ih2+high(1)*ih3
      out(2)=low(half)*ig0+high(half)*ig1+low(1)*ig2+high(1)*ig3
      j=3
      
      do i=1,half-1
   !     if(j==11) write(0,*) "I",j,i,low(i:i+1),out(j:j+1)
        out(j)=low(i)*ih0+high(i)*ih1+low(i+1)*ih2+high(i+1)*ih3
        out(j+1)=low(i)*ig0+high(i)*ig1+low(i+1)*ig2+high(i+1)*ig3
        j=j+2
      end do
     ! write(0,*) "LOOPING",half,low(11:14),high(11:14),out(11:14)

    end subroutine






end module
