module wavelet2_mod
   use adj_mod
  use sep_mod
   implicit none
   
   real, allocatable,private :: sm(:),rgh(:)
   integer,private :: levels(10),ns(10),l_sm,l_rgh
   logical,private :: tow
   contains
   subroutine init_wavelet_transform(smoother,roughner,n1,n2,n3,n4,lev1,lev2,lev3,lev4,ls,lr)
     integer :: n1,n2,n3,n4,lev1,lev2,lev3,lev4,ls,lr
     real    :: smoother(:),roughner(:)
     allocate(sm(size(smoother))); sm=smoother
     allocate(rgh(size(roughner))); rgh=roughner
     ns(1:4)=(/n1,n2,n3,n4/)
     levels(1:4)=(/lev1,lev2,lev3,lev4/)
     l_sm=ls
     l_rgh=lr
     tow=.true.
   end subroutine
   subroutine set_direction(dir)
     logical :: dir
     tow=dir
   end subroutine
 
    
    integer function wavelet_transform_op(adj,add,model,data)
      logical,intent(in) :: adj,add
      real               :: model(:),data(:)
      real,allocatable   :: m(:,:,:,:)
      integer            :: nlev,ilev,j1,j2,j3,j4
      integer            :: nj4,nj3,nj2,i4,i3,i2
      allocate(m(ns(1),ns(2),ns(3),ns(4)))
      wavelet_transform_op=0
      call adjnull(adj,add,model,data)
      nlev=maxval(levels)
      if(adj) then
        j1=1; j2=1; j3=1; j4=1;
        m=reshape(data,(/ns(1),ns(2),ns(3),ns(4)/))
        do ilev=1,nlev
          nj2=1; nj3=1; nj4=1
          if(levels(4)>=ilev) then
             call apply_filter_axis_4_2(adj,m,j1,j2,j3,j4,0,0,0,0)
             j4=j4*2; nj4=2
          end if
          if(levels(3)>=ilev) then
            do i4=0,nj4-1
              call apply_filter_axis_3_2(adj,m,j1,j2,j3,j4,0,0,0,i4*j4/2)
            end do
            j3=j3*2; nj3=2
          end if
          if(levels(2)>=ilev) then
            do i4=0,nj4-1
              do i3=0,nj3-1
                call apply_filter_axis_2_2(adj,m,j1,j2,j3,j4,0,0,i3*j3/2,i4*j4/2)
              end do
            end do
            j2=j2*2; nj2=2
          end if
          if(levels(1)>=ilev) then
            do i4=0,nj4-1
              do i3=0,nj3-1
                do i2=0,nj2-1
                  call apply_filter_axis_1_2(adj,m,j1,j2,j3,j4,0,i2*j2/2,i3*j3/2,i4*j4/2)
                end do
              end do
            end do
            j1=j1*2
          end if
        end do
        model=model+reshape(m,(/size(m)/))
      else
         m=reshape(model,(/ns(1),ns(2),ns(3),ns(4)/))
        j1=2**levels(1); j2=2**levels(2)
        j3=2**levels(3); j4=2**levels(4)
        do ilev=nlev,1,-1
          nj2=1;nj3=1; nj4=1;
          if(levels(2)>=ilev) nj2=2
          if(levels(3)>=ilev) nj3=2
          if(levels(4)>=ilev) nj4=2
          if(levels(1)>=ilev) then
            do i4=0,nj4-1
              do i3=0,nj3-1
                do i2=0,nj2-1
                  call apply_filter_axis_1_2(adj,m,j1,j2,j3,j4,0,i2*j2/4,i3*j3/4,i4*j4/4)
                end do
              end do
            end do
            j1=j1/2
          end if
          if(levels(2)>=ilev) then
            do i4=0,nj4-1
              do i3=0,nj3-1
                call apply_filter_axis_2_2(adj,m,j1,j2,j3,j4,0,0,i3*j3/4,i4*j4/4)
              end do
            end do
            j2=j2/2
          end if
          if(levels(3)>=ilev) then
            do i4=0,nj4-1
              call apply_filter_axis_3_2(adj,m,j1,j2,j3,j4,0,0,0,i4*j4/4)
            end do
            j3=j3/2
          end if
          if(levels(4)>=ilev) then
            call apply_filter_axis_4_2(adj,m,j1,j2,j3,j4,0,0,0,0)
            j4=j4/2
          end if        
        end do
        data=data+reshape(m,(/size(m)/))
      end if
      deallocate(m)
    end function
    
    subroutine apply_filter_1d(adj,m1,m2,data)
       real    :: data(:)
       real    :: m1(:),m2(:)
       logical,intent(in) :: adj
       integer :: i,i_f,i2,nh,i_2
       if(size(data)<size(sm) .or.size(data) < size(rgh)) call seperr("filter is larger than the data")
       if(1==1) then
       if(adj) then
          m1=0; m2=0;
          i2=1; i=1 ;do while(i2<=size(data))
            do i_f=1,size(sm)
                                   
                          i_2=mod(i2+i_f-1-l_sm+size(data),size(data))+1
              m1(i)=m1(i)+sm(i_f)*data(i_2)
            end do
                        i=i+1; i2=i2+2;
          end do
          i2=1; i=1 ;do while(i2<=size(data))
            do i_f=1,size(rgh)
                          i_2=mod(i2+i_f-1-l_rgh+size(data),size(data))+1

              m2(i)=m2(i)+rgh(i_f)*data(i_2)
            end do
            i=i+1; i2=i2+2;
          end do
       else
          data=0; 
          i2=1; i=1; do while(i2<=size(data))
            do i_f=1,size(sm)
                          i_2=mod(i2+i_f-1-l_sm+size(data),size(data))+1

              data(i_2)=m1(i)*sm(i_f)+data(i_2)
            end do
             i=i+1; i2=i2+2;
          end do
          
          !filter will always 
          i2=1; i=1; do while(i2<=size(data))
            do i_f=1,size(rgh)
                          i_2=mod(i2+i_f-1-l_rgh+size(data),size(data))+1

              data(i_2)=m2(i)*rgh(i_f)+data(i_2)
            end do
            i=i+1; i2=i2+2;
          end do
       end if
      end if
   
    end subroutine
 subroutine fwt97(adj,x)
   logical, intent(in) :: adj
   real                :: x(:)
   integer             :: i,n
   real                :: a
   n=size(x)
   if(adj) then
!if(1==3) then

 !    x=data
     a=-1.586134342;
     do i=2,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do

     x(n)=x(n)+2*a*x(n-1)
     a=-0.05298011854;

     do i=3,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do
     x(1)=x(1)+2*a*x(2)

     a=0.8829110762;
     do i=2,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do

     x(n)=x(n)+2*a*x(n-1)

     a=0.4435068522;
     do i=3,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do
     x(1)=x(1)+2*a*x(2)
 !      end if

     a=1/1.149604398;

     do i=1,n
       if(mod(i,2)==1) then
         x(i)=x(i)*a
       else
         x(i)=x(i)/a
       end if
     end do

 !    model=x+model
   else
  !   x=model

     a=1/1.149604398;
     do i=1,n
       if(mod(i,2)==1) then
         x(i)=x(i)*a
       else
         x(i)=x(i)/a
       end if
     end do
 ! if(1==3) then

     a=0.4435068522;
     x(2)=x(2)+2*a*x(1)
     do i=3,n-1,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do

     a=0.8829110762;
     x(n-1)=x(n-1)+2*a*x(n)
     do i=2,n-1,2
       x(i+1)=x(i+1)+a*x(i)
       x(i-1)=x(i-1)+a*x(i)
     end do
     a=-0.05298011854;
     x(2)=x(2)+2*a*x(1)
     do i=3,n-1,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do
     
     a=-1.586134342;
     x(n-1)=x(n-1)+2*a*x(n)
     do i=2,n-1,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do
 ! end if
 !    data=data+x
  end if
end subroutine
 
 subroutine iwt97(adj,x)
   logical, intent(in) :: adj
   real :: x(:)
   integer             :: i,n
   real                :: a
   n=size(x)
   if(.not. adj) then
     a=1.149604398;
     do i=1,n
       if(mod(i,2)==1) then
         x(i)=x(i)*a
       else
         x(i)=x(i)/a
       end if
     end do
     a=-0.4435068522;
     do i=3,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do

     x(1)=x(1)+2*a*x(2)

     a=-0.8829110762;
     do i=2,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do
     x(n)=x(n)+2*a*x(n-1)

     a=0.05298011854;
     do i=3,n,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do
     x(1)=x(1)+2*a*x(2)

     a=1.586134342;
     do i=2,n-3,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do
     x(n)=x(n)+2*a*x(n-1)
   else
     a=1.586134342;
     x(n-1)=x(n-1)+2*a*x(n)
     do i=2,n-3,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do

     a=0.05298011854;
     x(2)=x(2)+2*a*x(1)
     do i=3,n,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do

     a=-0.8829110762;
     x(n-1)=x(n-1)+2*a*x(n)
     do i=2,n-1,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do

     a=-0.4435068522;
     x(2)=x(2)+x(1)*2*a
     do i=3,n-1,2
       x(i+1)=x(i+1)+a*x(i)
       x(i-1)=x(i-1)+a*x(i)
     end do
     a=1.149604398;
     do i=1,n
       if(mod(i,2)==1) then
         x(i)=x(i)*a
       else
         x(i)=x(i)/a
       end if
     end do
   end if
 end subroutine
      subroutine apply_filter_axis_4_2(adj,data,j1,j2,j3,j4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,i3,i4,i1,ns,nsh,i,f1,f2,f3,f4,j1,j2,j3,j4
      integer            :: n1,n2,n3,n4,j_2,j_3,j_1
      real,allocatable   :: td(:),tm(:)
      n1=size(data,1); n2=size(data,2);
      n3=size(data,3); n4=size(data,4);
      if(.not. adj) then
        nsh=ceiling(size(data,4)*1./j4*1.)
        ns=nsh*2
        j_2=max(1,j2/2); j_3=max(1,j3/2); j_1=max(1,j1/2)
      else
        ns=ceiling(size(data,4)*1./j4*1.)
        j_2=j2; j_3=j3; j_1=j1
        nsh=ceiling(ns*1./2.)      
      end if
              j_2=j2; j_3=j3; j_1=j1

      allocate(td(ns),tm(ns))
      n3=1; n2=1; n1=1
      write(0,*) "TIME TO LOOP",adj,f1,j_1,f4,j4
      do i3=1+f3,n3,j_3
        do i2=1+f2,n2,j_2
          do i1=1+f1,n1,j_1
            if(adj) then !to wavelet 
              !copy in current level
              i=1
              do i4=1+f4,n4,j4
                td(i)=data(i1,i2,i3,i4); i=i+1
              end do
              if(tow) then
                call fwt97(adj,td)
              else
                call iwt97(adj,td)
              end if
              !put the data back in
              i=1
              do i4=1+f4,n4,j4
                data(i1,i2,i3,i4)=td(i); i=i+1
              end do
            else ! from wavelet
              i=1
              do i4=1+f4,n4,j4/2
                tm(i)=data(i1,i2,i3,i4); i=i+1
              end do
              i=1
              if(tow) then
                call fwt97(adj,tm)
              else
                call iwt97(adj,tm)
              end if
              !copy in current level
              i=1
              do i4=1+f4,n4,j4/2
                data(i1,i2,i3,i4)=tm(i); i=i+1
              end do
            end if
          end do
        end do
      end do
      deallocate(td,tm)
    end subroutine
      subroutine apply_filter_axis_3_2(adj,data,j1,j2,j3,j4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,i3,i4,i1,ns,nsh,i
      integer            :: n1,n2,n3,n4,f1,f2,f3,f4,j1,j2,j3,j4,j_2,j_1,j_4
      real,allocatable   :: td(:),tm(:)
      n1=size(data,1); n2=size(data,2);
      n3=size(data,3); n4=size(data,4);
      if(.not. adj) then
        nsh=ceiling(size(data,3)*1./j3*1.)
        ns=nsh*2
         j_2=max(1,j2/2); j_1=max(1,j1/2); j_4=max(1,j4/2)

      else
        ns=ceiling(size(data,3)*1./j3*1.)
        nsh=ceiling(ns*1./2.)     
        j_2=j2; j_4=j4; j_1=j1
      end if
      allocate(td(ns),tm(ns))
      n1=1; n2=1; n4=1;
      do i4=1+f4,n4,j_4
        do i2=1+f2,n2,j_2
          do i1=1+f1,n1,j_1
            if(adj) then !to wavelet 
              !copy in current level
              i=1
              do i3=1+f3,n3,j3
                td(i)=data(i1,i2,i3,i4); i=i+1
              end do
              if(tow) then
                call fwt97(adj,td)
              else
                 call fwt97(adj,td)
              end if
              !put the data back in
              i=1
              do i3=1+f3,n3,j3
                data(i1,i2,i3,i4)=td(i); i=i+1
              end do
            else ! from wavelet
              i=1
              do i3=1+f3,n3,j3/2
                td(i)=data(i1,i2,i3,i4); i=i+1
              end do
              if(tow) then
                call fwt97(adj,td)
              else
                call iwt97(adj,td)
              end if
              !copy in current level
              i=1
              do i3=1+f3,n3,j3/2
                data(i1,i2,i3,i4)=td(i); i=i+1
              end do
            end if
          end do
        end do
      end do
      deallocate(td,tm)
    end subroutine
    
    
    
    subroutine apply_filter_axis_2_2(adj,data,j1,j2,j3,j4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,i3,i4,i1,ns,nsh,i,j1,j2,j3,j4,f1,f2,f3,f4
      integer            :: n1,n2,n3,n4,j_1,j_3,j_4
      real,allocatable   :: td(:),tm(:)
      n1=size(data,1); n2=size(data,2);
      n3=size(data,3); n4=size(data,4);
      if(.not. adj) then
        nsh=ceiling(size(data,2)*1./j2*1.)
        ns=nsh*2
       j_1=max(1,j1/2); j_3=max(1,j3/2); j_4=max(1,j4/2)

      else
        ns=ceiling(size(data,2)*1./j2*1.)
        nsh=ceiling(ns*1./2.)  
        j_1=j1; j_4=j4; j_3=j3
      end if
      allocate(td(ns),tm(ns))
      n4=1; n3=1; n1=1
      do i4=1+f4,n4,j_4
        do i3=1+f3,n3,j_3
          do i1=1+f1,n1,j_1
            if(adj) then !to wavelet 
              !copy in current level
              i=1
              do i2=1+f2,n2,j2
                td(i)=data(i1,i2,i3,i4); i=i+1
              end do
              if(tow) then
                call fwt97(adj,td)
              else
                call iwt97(adj,td)
              end if
              !put the data back in
              i=1
             do i2=1+f2,n2,j2
               data(i1,i2,i3,i4)=td(i); i=i+1
             end do

            else ! from wavelet
              i=1
             do i2=1+f2,n2,j2/2
               tm(i)=data(i1,i2,i3,i4); i=i+1
             end do
             if(tow) then
              call fwt97(adj,tm)
             else
               call iwt97(adj,tm)
             end if
              !copy in current level
        i=1
             do i2=1+f2,n2,j2/2
       data(i1,i2,i3,i4)=tm(i);i=i+1
             end do
            end if
          end do
        end do
      end do
      deallocate(td,tm)
    end subroutine





    subroutine apply_filter_axis_1_2(adj,data,j1,j2,j3,j4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,i3,i4,i1,ns,nsh,i
      integer            :: n1,n2,n3,n4,f1,f2,f3,f4,j1,j2,j3,j4
      integer            :: j_2,j_3,j_4
      real,allocatable   :: td(:),tm(:)
      n1=size(data,1); n2=size(data,2);
      n3=size(data,3); n4=size(data,4);
      if(.not. adj) then
        nsh=ceiling(size(data,1)*1./j1*1.)
        ns=nsh*2
        j_2=max(1,j2/2); j_3=max(1,j3/2); j_4=max(1,j4/2)
      else
        ns=ceiling(size(data,1)*1./j1*1.)
        nsh=ceiling(ns*1./2.)      
        j_2=j2; j_3=j3; j_4=j4;
      end if
      allocate(td(ns),tm(ns))
      do i4=1+f4,n4,j_4
        do i3=1+f3,n3,j_3
          do i2=1+f2,n2,j_2
   
            if(adj) then !to wavelet 
              !copy in current level
              i=1
              do i1=1+f1,n1,j1
               td(i)=data(i1,i2,i3,i4); i=i+1
              end do
              if(tow) then 
                call fwt97(adj,td)
              else
                call iwt97(adj,td)
              end if
              !put the data back in
              i=1
              do i1=1+f1,n1,j1
                data(i1,i2,i3,i4)=td(i); i=i+1
              end do    
            else ! from wavelet
              i=1
              do i1=1+f1,n1,j1/2
                tm(i)=data(i1,i2,i3,i4); i=i+1
              end do
              i=1
              if(tow) then 
                call fwt97(adj,tm)
              else
                call iwt97(adj,tm)
              end if
              !copy in current level
              i=1
              do i1=1+f1,n1,j1/2
                data(i1,i2,i3,i4)=tm(i); i=i+1
              end do
            end if
          end do
        end do
      end do
      deallocate(td,tm)
    end subroutine
            
 
 
     subroutine apply_filter_axis_4(adj,data,j1,j2,j3,j4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,i3,i4,i1,ns,nsh,i,f1,f2,f3,f4,j1,j2,j3,j4
      integer            :: n1,n2,n3,n4
      real,allocatable   :: td(:),tm1(:),tm2(:)
      n1=size(data,1); n2=size(data,2);
      n3=size(data,3); n4=size(data,4);
      if(.not. adj) then
        nsh=ceiling(size(data,4)*1./j4*1.)
        ns=nsh*2
      else
        ns=ceiling(size(data,4)*1./j4*1.)
        nsh=ceiling(ns*1./2.)      
      end if
      allocate(td(ns),tm1(nsh),tm2(nsh))
      n3=1; n2=1; n1=1
      do i3=1+f3,n3,j3
        do i2=1+f2,n2,j2
          do i1=1+f1,n1,j1
            if(adj) then !to wavelet 
              !copy in current level
              i=1
              do i4=1+f4,n4,j4
                td(i)=data(i1,i2,i3,i4); i=i+1
              end do
              call apply_filter_1d(adj,tm1,tm2,td)
              !put the data back in
              i=1
              do i4=1+f4,n4,j4*2
                data(i1,i2,i3,i4)=tm1(i); i=i+1
              end do
              i=1
              do i4=1+j4+f4,n4,j4*2
                data(i1,i2,i3,i4)=tm2(i); i=i+1
              end do
            else ! from wavelet
              i=1
              do i4=1+f4,n4,j4
                tm1(i)=data(i1,i2,i3,i4); i=i+1
              end do
              i=1
              do i4=1+f4+j4/2,n4,j4
                tm2(i)=data(i1,i2,i3,i4); i=i+1
              end do
              call apply_filter_1d(adj,tm1,tm2,td)
              !copy in current level
              i=1
              do i4=1+f4,n4,j4/2
                data(i1,i2,i3,i4)=td(i); i=i+1
              end do
            end if
          end do
        end do
      end do
      deallocate(td,tm1,tm2)
    end subroutine
      subroutine apply_filter_axis_3(adj,data,j1,j2,j3,j4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,i3,i4,i1,ns,nsh,i
      integer            :: n1,n2,n3,n4,f1,f2,f3,f4,j1,j2,j3,j4
      real,allocatable   :: td(:),tm1(:),tm2(:)
      n1=size(data,1); n2=size(data,2);
      n3=size(data,3); n4=size(data,4);
      if(.not. adj) then
        nsh=ceiling(size(data,3)*1./j3*1.)
        ns=nsh*2
      else
        ns=ceiling(size(data,3)*1./j3*1.)
        nsh=ceiling(ns*1./2.)      
      end if
      allocate(td(ns),tm1(nsh),tm2(nsh))
      n1=1; n2=1; n4=1;
      do i4=1+f4,n4,j4
        do i2=1+f2,n2,j2
          do i1=1+f1,n1,j1
            if(adj) then !to wavelet 
              !copy in current level
              i=1
              do i3=1+f3,n3,j3
                td(i)=data(i1,i2,i3,i4); i=i+1
              end do
              call apply_filter_1d(adj,tm1,tm2,td)
              !put the data back in
              i=1
              do i3=1+f3,n3,j3*2
                data(i1,i2,i3,i4)=tm1(i); i=i+1
              end do
              i=1
              do i3=1+j3+f3,n3,j3*2
                data(i1,i2,i3,i4)=tm2(i); i=i+1
              end do
            else ! from wavelet
              i=1
              do i3=1+f3,n3,j3
                tm1(i)=data(i1,i2,i3,i4); i=i+1
              end do
              i=1
              do i3=1+j3/2+f3,n3,j3
                tm2(i)=data(i1,i2,i3,i4); i=i+1
              end do
              call apply_filter_1d(adj,tm1,tm2,td)
              !copy in current level
              i=1
              do i3=1+f3,n3,j3/2
                data(i1,i2,i3,i4)=td(i); i=i+1
              end do
            end if
          end do
        end do
      end do
      deallocate(td,tm1,tm2)
    end subroutine
    
    
    
    subroutine apply_filter_axis_2(adj,data,j1,j2,j3,j4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,i3,i4,i1,ns,nsh,i,j1,j2,j3,j4,f1,f2,f3,f4
      integer            :: n1,n2,n3,n4
      real,allocatable   :: td(:),tm1(:),tm2(:)
      n1=size(data,1); n2=size(data,2);
      n3=size(data,3); n4=size(data,4);
      if(.not. adj) then
        nsh=ceiling(size(data,2)*1./j2*1.)
        ns=nsh*2
      else
        ns=ceiling(size(data,2)*1./j2*1.)
        nsh=ceiling(ns*1./2.)      
      end if
      allocate(td(ns),tm1(nsh),tm2(nsh))
      n4=1; n3=1; n1=1
      do i4=1+f4,n4,j4
        do i3=1+f3,n3,j3
          do i1=1+f1,n1,j1
            if(adj) then !to wavelet 
              !copy in current level
              i=1
              do i2=1+f2,n2,j2
                td(i)=data(i1,i2,i3,i4); i=i+1
              end do
              call apply_filter_1d(adj,tm1,tm2,td)
              !put the data back in
              i=1
              do i2=1+f2,n2,j2*2
                data(i1,i2,i3,i4)=tm1(i); i=i+1
              end do
              i=1
              do i2=1+j2+f2,n2,j2*2
                data(i1,i2,i3,i4)=tm2(i); i=i+1
              end do
            else ! from wavelet
              i=1
              do i2=1+f2,n2,j2
                tm1(i)=data(i1,i2,i3,i4); i=i+1
              end do
              i=1
              do i2=1+j2/2+f2,n2,j2
                tm2(i)=data(i1,i2,i3,i4); i=i+1
              end do
              call apply_filter_1d(adj,tm1,tm2,td)
              !copy in current level
              i=1
              do i2=1+f2,n2,j2/2
                data(i1,i2,i3,i4)=td(i);i=i+1
              end do
            end if
          end do
        end do
      end do
      deallocate(td,tm1,tm2)
    end subroutine





    subroutine apply_filter_axis_1(adj,data,j1,j2,j3,j4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,i3,i4,i1,ns,nsh,i
      integer            :: n1,n2,n3,n4,f1,f2,f3,f4,j1,j2,j3,j4
      real,allocatable   :: td(:),tm1(:),tm2(:)
      n1=size(data,1); n2=size(data,2);
      n3=size(data,3); n4=size(data,4);
      if(.not. adj) then
        nsh=ceiling(size(data,1)*1./j1*1.)
        ns=nsh*2
      else
        ns=ceiling(size(data,1)*1./j1*1.)
        nsh=ceiling(ns*1./2.)      
      end if

      allocate(td(ns),tm1(nsh),tm2(nsh))

      do i4=1+f4,n4,j4
        do i3=1+f3,n3,j3
          do i2=1+f2,n2,j2
            if(adj) then !to wavelet 
              !copy in current level
              i=1
              if(1==1) then
              do i1=1+f1,n1,j1
                td(i)=data(i1,i2,i3,i4); i=i+1
              end do
              call apply_filter_1d(adj,tm1,tm2,td)
              !put the data back in
              i=1
              do i1=1+f1,n1,j1*2
                data(i1,i2,i3,i4)=tm1(i); i=i+1
              end do
              i=1
              do i1=1+j1+f1,n1,j1*2
                data(i1,i2,i3,i4)=tm2(i); i=i+1
              end do
              else
              td=data(1:32,i2,i3,i4)
              call apply_filter_1d(adj,tm1,tm2,td)
              data(1:16,i2,i3,i4)=tm1
              data(17:32,i2,i3,i4)=tm2
          !    data(1:31:2,i2,i3,i4)=tm1
           !   data(2:32:2,i2,i3,i4)=tm2
              end if
              
            else ! from wavelet
            if(1==1) then
              i=1
              do i1=1+f1,n1,j1
                tm1(i)=data(i1,i2,i3,i4); i=i+1
              end do
              i=1
              do i1=1+j1/2+f1,n1,j1
                tm2(i)=data(i1,i2,i3,i4); i=i+1
              end do
              call apply_filter_1d(adj,tm1,tm2,td)
              !copy in current level
              i=1
              do i1=1+f1,n1,j1/2
                data(i1,i2,i3,i4)=td(i); i=i+1
              end do
             else
          !   tm1=data(1:31:2,i2,i3,i4)
          !   tm2=data(2:32:2,i2,i3,i4)
             tm1=data(1:16,i2,i3,i4)
             tm2=data(17:32,i2,i3,i4)
             call apply_filter_1d(adj,tm1,tm2,td)
             data(1:32,i2,i3,i4)=td
             endif
            end if
          end do
        end do
      end do
      deallocate(td,tm1,tm2)
    end subroutine
            
    
    

    
    




end module
