module wavelet3_mod
   use sep_mod
   use adj_mod
   implicit none
   real, allocatable,private :: sm(:),rgh(:)
   integer,private :: levels(10),ns(10)
   logical,private :: tow
   contains
   subroutine init_wavelet_transform(smoother,roughner,n1,n2,n3,n4,lev1,lev2,lev3,lev4)
     integer :: n1,n2,n3,n4,lev1,lev2,lev3,lev4,ls,lr
     real    :: smoother(:),roughner(:)
     allocate(sm(size(smoother))); sm=smoother
     allocate(rgh(size(roughner))); rgh=roughner
     ns(1:4)=(/n1,n2,n3,n4/)
     levels(1:4)=(/lev1,lev2,lev3,lev4/)
     tow=.true.
     if(n1/2**lev1 < 5) call seperr("to many levels for axis 1")
     if(n2/2**lev2 < 5) call seperr("to many levels for axis 2")
     if(n3/2**lev3 < 5) call seperr("to many levels for axis 3")
     if(n4/2**lev4 < 5) call seperr("to many levels for axis 4")

   end subroutine
   subroutine set_direction(dir)
     logical :: dir
     tow=dir
   end subroutine
 
    subroutine to_wavelet(input,output)
      real :: input(:),output(:)
      integer :: ierr
      call set_direction(.true.)
      ierr=wavelet_transform_op(.true.,.false.,output,input)
    end subroutine
    subroutine from_wavelet(input,output)
      real :: input(:),output(:)
      integer :: ierr
      call set_direction(.false.)
      ierr=wavelet_transform_op(.false.,.false.,input,output)
    end subroutine
    integer function wavelet_transform_op(adj,add,model,data)
      logical,intent(in) :: adj,add
      real               :: model(:),data(:)
      real,allocatable   :: m(:,:,:,:)
      integer            :: nlev,ilev,j1,j2,j3,j4
      integer            :: nj4,nj3,nj2,i4,i3,i2,n1,n2,n3,n4
      allocate(m(ns(1),ns(2),ns(3),ns(4)))
      wavelet_transform_op=0
      call adjnull(adj,add,model,data)
      nlev=maxval(levels)
      n1=ns(1); n2=ns(2); n3=ns(3); n4=ns(4)
      if(adj) then
        m=reshape(data,(/ns(1),ns(2),ns(3),ns(4)/))
        do ilev=1,nlev
          nj2=1; nj3=1; nj4=1
          if(levels(4)>=ilev) then
             call apply_filter_axis_4(adj,m,n1,n2,n3,n4,0,0,0,0)
             n4=n4/2; nj4=2
          end if
          if(levels(3)>=ilev) then
            do i4=0,nj4-1
              call apply_filter_axis_3(adj,m,n1,n2,n3,n4,0,0,0,n4*i4)
            end do
            n3=n3/2; nj3=2
          end if
          if(levels(2)>=ilev) then
            do i4=0,nj4-1
              do i3=0,nj3-1
                call apply_filter_axis_2(adj,m,n1,n2,n3,n4,0,0,n3*i3,n4*i4)
              end do
            end do
            n2=n2/2; nj2=2
          end if
          if(levels(1)>=ilev) then
            do i4=0,nj4-1
              do i3=0,nj3-1
                do i2=0,nj2-1
                  call apply_filter_axis_1(adj,m,n1,n2,n3,n4,0,n2*i2,n3*i3,n4*i4)
                end do
              end do
            end do
            n1=n1/2
          end if
        end do
        model=model+reshape(m,(/size(m)/))
      else
         m=reshape(model,(/ns(1),ns(2),ns(3),ns(4)/))
         n1=ns(1)/2**levels(1); n2=ns(2)/2**levels(2)
         n3=ns(3)/2**levels(3); n4=ns(4)/2**levels(4)
        do ilev=nlev,1,-1
          nj2=1;nj3=1; nj4=1;
          if(levels(2)>=ilev) nj2=2
          if(levels(3)>=ilev) nj3=2
          if(levels(4)>=ilev) nj4=2
          if(levels(1)>=ilev) then
           n1=n1*2
            do i4=0,nj4-1
              do i3=0,nj3-1
                do i2=0,nj2-1
                  call apply_filter_axis_1(adj,m,n1,n2,n3,n4,0,n2*i2,n3*i3,n4*i4)
                end do
              end do
            end do
          end if
          if(levels(2)>=ilev) then
            n2=n2*2
            do i4=0,nj4-1
              do i3=0,nj3-1
                call apply_filter_axis_2(adj,m,n1,n2,n3,n4,0,0,n3*i3,n4*i4)
              end do
            end do
          end if
          if(levels(3)>=ilev) then
            n3=n3*2
            do i4=0,nj4-1
              call apply_filter_axis_3(adj,m,n1,n2,n3,n4,0,0,0,n4*i4)
            end do
          end if
          if(levels(4)>=ilev) then
            n4=n4*2
            call apply_filter_axis_4(adj,m,n1,n2,n3,n4,0,0,0,0)
          end if        
        end do
        data=data+reshape(m,(/size(m)/))
      end if
      deallocate(m)
    end function
    
    
 subroutine fwt97(adj,x)
   logical, intent(in) :: adj
   real                :: x(:)
   integer             :: i,n
   real                :: a
   n=size(x)
   if(adj) then
!if(1==3) then

 !    x=data
     a=-1.586134342;
     do i=2,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do

     x(n)=x(n)+2*a*x(n-1)
     a=-0.05298011854;

     do i=3,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do
     x(1)=x(1)+2*a*x(2)

     a=0.8829110762;
     do i=2,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))

     end do


     x(n)=x(n)+2*a*x(n-1)

     a=0.4435068522;
     do i=3,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))

     end do
     x(1)=x(1)+2*a*x(2)
 !      end if
     a=1/1.149604398;

     do i=1,n
       if(mod(i,2)==0) then
         x(i)=x(i)*a
       else
         x(i)=x(i)/a
       end if

     end do
 !    model=x+model
   else
  !   x=model

     a=1/1.149604398;
     do i=1,n
       if(mod(i,2)==0) then
         x(i)=x(i)*a
       else
         x(i)=x(i)/a
       end if
     end do
 ! if(1==3) then

     a=0.4435068522;
     x(2)=x(2)+2*a*x(1)
     do i=3,n-1,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do

     a=0.8829110762;
     x(n-1)=x(n-1)+2*a*x(n)
     do i=2,n-1,2
       x(i+1)=x(i+1)+a*x(i)
       x(i-1)=x(i-1)+a*x(i)
     end do
     a=-0.05298011854;
     x(2)=x(2)+2*a*x(1)
     do i=3,n-1,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do
     
     a=-1.586134342;
     x(n-1)=x(n-1)+2*a*x(n)
     do i=2,n-1,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do
 ! end if
 !    data=data+x
  end if
end subroutine
 
 subroutine iwt97(adj,x)
   logical, intent(in) :: adj
   real :: x(:)
   integer             :: i,n
   real                :: a
   n=size(x)
   if(.not. adj) then
     a=1.149604398;
     do i=1,n
       if(mod(i,2)==0) then
         x(i)=x(i)*a
       else
         x(i)=x(i)/a
       end if
     end do

     a=-0.4435068522;
     do i=3,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do

     x(1)=x(1)+2*a*x(2)
     a=-0.8829110762;
     do i=2,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do
     x(n)=x(n)+2*a*x(n-1)

     a=0.05298011854;
     do i=3,n,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do
     x(1)=x(1)+2*a*x(2)
!if(1==3) then
     a=1.586134342;
     do i=2,n-1,2
       x(i)=x(i)+a*(x(i-1)+x(i+1))
     end do
 x(n)=x(n)+2*a*x(n-1)
   else
     a=1.586134342;
     x(n-1)=x(n-1)+2*a*x(n)
     do i=2,n-3,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do

     a=0.05298011854;
     x(2)=x(2)+2*a*x(1)
     do i=3,n,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do

     a=-0.8829110762;
     x(n-1)=x(n-1)+2*a*x(n)
     do i=2,n-1,2
       x(i-1)=x(i-1)+a*x(i)
       x(i+1)=x(i+1)+a*x(i)
     end do

     a=-0.4435068522;
     x(2)=x(2)+x(1)*2*a
     do i=3,n-1,2
       x(i+1)=x(i+1)+a*x(i)
       x(i-1)=x(i-1)+a*x(i)
     end do
     a=1.149604398;
     do i=1,n
       if(mod(i,2)==1) then
         x(i)=x(i)*a
       else
         x(i)=x(i)/a
       end if
     end do
     
   end if
 end subroutine
      subroutine apply_filter_axis_4(adj,data,n1,n2,n3,n4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,nth,ith,i3,i4,i1,nh,n1,n2,n3,n4,f1,f2,f3,f4
      integer, external :: omp_get_num_threads,omp_get_thread_num
      real,allocatable   :: x(:,:)
      !$OMP PARALLEL
      nth=omp_get_num_threads()
      !$OMP END PARALLEL 
      allocate(x(n4,nth))
      nh=n4/2
      if(adj) then
        !$OMP PARALLEL DO private(ith,i1,i2,i3,i4)
        do i3=1+f3,n3+f3
                  ith=omp_get_thread_num()+1
          do i2=1+f2,n2+f2
            do i1=1+f1,n1+f1
              x(:,ith)=data(i1,i2,i3,f4+1:f4+n4)
              if(tow) then
                call fwt97(adj,x(:,ith))
              else
                call iwt97(adj,x(:,ith))
              end if
              do i4=0,n4-1
                if(mod(i4,2)==0) then
                  data(i1,i2,i3,i4/2+1)=x(i4+1,ith)
                else
                  data(i1,i2,i3,i4/2+nh+1)=x(i4+1,ith)
                end if
              end do
            end do
          end do
        end do
                !OMP END PARALLEL DO

      else
              !$OMP PARALLEL DO private(ith,i1,i2,i3,i4)

        do i3=1+f3,n3+f3
                  ith=omp_get_thread_num()+1

          do i2=1+f2,n2+f2
            do i1=1+f1,n1+f1
              do i4=0,n4-1
                if(mod(i4,2)==0) then
                  x(i4+1,ith)=data(i1,i2,i3,i4/2+1)
                else
                  x(i4+1,ith)=data(i1,i2,i3,i4/2+nh+1)
                end if
              end do
              if(tow) then
                call fwt97(adj,x(:,ith))
              else
                call iwt97(adj,x(:,ith))
              end if
              data(i1,i2,i3,f4+1:f4+n4)=x(:,ith)

            end do
          end do
        end do
                !OMP END PARALLEL DO

      end if
     deallocate(x)
  end subroutine
      

 subroutine apply_filter_axis_3(adj,data,n1,n2,n3,n4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,i3,ith,nth,i4,i1,nh,n1,n2,n3,n4,f1,f2,f3,f4
      real,allocatable   :: x(:,:)
            integer, external :: omp_get_num_threads,omp_get_thread_num
      !$OMP PARALLEL
      nth=omp_get_num_threads()
      !$OMP END PARALLEL 
      allocate(x(n3,nth))
      nh=n3/2
      if(adj) then
              !$OMP PARALLEL DO private(ith,i1,i2,i3,i4)

        do i4=1+f4,n4+f4
                  ith=omp_get_thread_num()+1

          do i2=1+f2,n2+f2
            do i1=1+f1,n1+f1
              x(:,ith)=data(i1,i2,f3+1:f3+n3,i4)
              if(tow) then
                call fwt97(adj,x(:,ith))
              else
                call iwt97(adj,x(:,ith))
              end if
              do i3=0,n3-1
                if(mod(i3,2)==0) then
                  data(i1,i2,i3/2+1,i4)=x(i3+1,ith)
                else
                  data(i1,i2,i3/2+nh+1,i4)=x(i3+1,ith)
                end if
              end do
            end do
          end do
        end do
                !OMP END PARALLEL DO

      else
              !$OMP PARALLEL DO private(ith,i1,i2,i3,i4)

        do i4=1+f4,n4+f4
                  ith=omp_get_thread_num()+1

          do i2=1+f2,n2+f2
            do i1=1+f1,n1+f1
              do i3=0,n3-1
                if(mod(i3,2)==0) then
                  x(i3+1,ith)=data(i1,i2,i3/2+1,i4)
                else
                  x(i3+1,ith)=data(i1,i2,i3/2+nh+1,i4)
                end if
              end do
              if(tow) then
                call fwt97(adj,x(:,ith))
              else
                call iwt97(adj,x(:,ith))
              end if
              data(i1,i2,f3+1:f3+n3,i4)=x(:,ith)
            end do
          end do
        end do
                !OMP END PARALLEL DO

      end if
     deallocate(x)
  end subroutine
   subroutine apply_filter_axis_2(adj,data,n1,n2,n3,n4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: i2,nth,i3,ith,i4,i1,nh,n1,n2,n3,n4,f1,f2,f3,f4
      real,allocatable   :: x(:,:)
            integer, external :: omp_get_num_threads,omp_get_thread_num
      !$OMP PARALLEL
      nth=omp_get_num_threads()
      !$OMP END PARALLEL 
      allocate(x(n2,nth))
      nh=n2/2
      if(adj) then
              !$OMP PARALLEL DO private(ith,i1,i2,i3,i4)

        do i4=1+f4,n4+f4
                  ith=omp_get_thread_num()+1

          do i3=1+f3,n3+f3
            do i1=1+f1,n1+f1
              x(:,ith)=data(i1,f2+1:f2+n4,i3,i4)
              if(tow) then
                call fwt97(adj,x(:,ith))
              else
                call iwt97(adj,x(:,ith))
              end if
              do i2=0,n2-1
                if(mod(i2,2)==0) then
                  data(i1,i2/2+1,i3,i4)=x(i2+1,ith)
                else
                  data(i1,i2/2+1+nh,i3,i4)=x(i2+1,ith)
                end if
              end do
            end do
          end do
        end do
      else
              !$OMP PARALLEL DO private(ith,i1,i2,i3,i4)

        do i4=1+f4,n4+f4
                  ith=omp_get_thread_num()+1

          do i3=1+f3,n3+f3
            do i1=1+f1,n1+f1
              do i2=0,n2-1
                if(mod(i2,2)==0) then
                  x(i2+1,ith)=data(i1,i2/2+1,i3,i4)
                else
                  x(i2+1,ith)=data(i1,i2/2+1+nh,i3,i4)
                end if
              end do
              if(tow) then
                call fwt97(adj,x(:,ith))
              else
                call iwt97(adj,x(:,ith))
              end if
              data(i1,f2+1:f2+n2,i3,i4)=x(:,ith)

            end do
          end do
        end do
                !OMP END PARALLEL DO

      end if
     deallocate(x)
  end subroutine
   subroutine apply_filter_axis_1(adj,data,n1,n2,n3,n4,f1,f2,f3,f4)
      logical,intent(in) :: adj
      real               :: data(:,:,:,:)
      integer            :: ith,i2,nth,i3,i4,i1,nh,n1,n2,n3,n4,f1,f2,f3,f4
      real,allocatable   :: x(:,:)
            integer, external :: omp_get_num_threads,omp_get_thread_num
      !$OMP PARALLEL
      nth=omp_get_num_threads()
      !$OMP END PARALLEL 

      allocate(x(n1,nth))
      nh=n1/2
      if(adj) then
              !$OMP PARALLEL DO private(ith,i1,i2,i3,i4)

        do i4=1+f4,n4+f4
                  ith=omp_get_thread_num()+1

          do i3=1+f3,n3+f3
            do i2=1+f2,n2+f2
              x(:,ith)=data(f1+1:f1+n1,i2,i3,i4)
              if(tow) then
                call fwt97(adj,x(:,ith))
              else
                call iwt97(adj,x(:,ith))
              end if
              do i1=0,n1-1
                if(mod(i1,2)==0) then
                  data(i1/2+1,i2,i3,i4)=x(i1+1,ith)
                else
                  data(i1/2+nh+1,i2,i3,i4)=x(i1+1,ith)
                end if
              end do
            end do
          end do
        end do
        !$OMP END PARALLEL DO
      else
              !$OMP PARALLEL DO private(ith,i1,i2,i3,i4)

        do i4=1+f4,n4+f4
                  ith=omp_get_thread_num()+1

          do i3=1+f3,n3+f3
            do i2=1+f2,n2+f2
              do i1=0,n1-1
                if(mod(i1,2)==0) then
                  x(i1+1,ith)=data(i1/2+1,i2,i3,i4)
                else
                  x(i1+1,ith)=data(i1/2+nh+1,i2,i3,i4)
                end if
              end do
              if(tow) then
                call fwt97(adj,x(:,ith))
              else
                call iwt97(adj,x(:,ith))
              end if
              data(1+f1:f1+n1,i2,i3,i4)=x(:,ith)

            end do
          end do
        end do
        !OMP END PARALLEL DO
      end if
     deallocate(x)
  end subroutine



end module
